# `name` is the name of the package as used for `pip install package`
name = "frarch"
# `path` is the name of the package for `import package`
path = name.lower().replace("-", "_").replace(" ", "_")
# Version number should follow https://python.org/dev/peps/pep-0440 and
# https://semver.org
version = "0.1.4"
author = "vbadenas"
author_email = "victor.badenas@gmail.com"
description = "Training Framework for PyTorch projects"  # One-liner
url = "https://github.com/victorbadenas/frarch"  # your project homepage
license = "Apache 2.0"  # See https://choosealicense.com



"""
class definition of a base trainer.

:Description: base trainer class

:Authors: victor badenas (victor.badenas@gmail.com)

:Version: 0.1.0
"""

__title__ = "base_trainer"
__version__ = "0.1.0"
__author__ = "victor badenas"

import logging
import sys
from typing import Any, Mapping, Optional, Type, Union

import torch
from torch.utils.data import DataLoader, Dataset

from frarch.modules import Checkpointer
from frarch.utils.stages import Stage

logger = logging.getLogger(__name__)
PYTHON_VERSION_MAJOR = 3
PYTHON_VERSION_MINOR = 7

default_values = {
    "debug": False,
    "debug_batches": 2,
    "device": "cpu",
    "nonfinite_patience": 3,
    "noprogressbar": False,
    "ckpt_interval_minutes": None,
    "train_interval": 10,
}


class BaseTrainer:
    """Abstract class for trainer managers.

    Args:
        modules (Mapping[str, torch.nn.Module]): trainable modules in the training.
        opt_class (Type[torch.optim.Optimizer]): optimizer class for training.
        hparams (Mapping[str, Any]): hparams dict-like structure from hparams file.
        checkpointer (Optional[Checkpointer], optional): Checkpointer class for saving
            the model and the hyperparameters needed. If None, no checkpoints are saved.
            Defaults to None.

    Raises:
        ValueError: ckpt_interval_minutes must be > 0 or None
        SystemError: Python version not supported. Python version must be >= 3.7
    """

    def __init__(
        self,
        modules: Mapping[str, torch.nn.Module],
        opt_class: Type[torch.optim.Optimizer],
        hparams: Mapping[str, Any],
        checkpointer: Optional[Checkpointer] = None,
    ) -> None:
        self.hparams = hparams
        self.opt_class = opt_class
        self.checkpointer = checkpointer

        for name, value in default_values.items():
            if name in self.hparams:
                if value != hparams[name]:
                    logger.info(
                        f"Parameter {name} overriden from"
                        f" default value and set to {hparams[name]}"
                    )
                setattr(self, name, hparams[name])
            else:
                setattr(self, name, value)

        if self.ckpt_interval_minutes is not None:
            if self.ckpt_interval_minutes <= 0:
                raise ValueError("ckpt_interval_minutes must be > 0 or None")

        # Check Python version
        if not (
            sys.version_info.major == PYTHON_VERSION_MAJOR
            and sys.version_info.minor >= PYTHON_VERSION_MINOR
        ):
            msg = (
                f"Detected Python {sys.version_info.major}.{sys.version_info.minor}."
                f"Python >= {PYTHON_VERSION_MAJOR}.{PYTHON_VERSION_MINOR} is required"
            )
            logger.error(msg)
            raise SystemError(msg)

        self.modules = torch.nn.ModuleDict(modules).to(self.device)

        # Prepare iterating variables
        self.avg_train_loss = 0.0
        self.step = 0

    def __call__(self, *args, **kwargs):
        """Alias for fit."""
        return self.fit(*args, **kwargs)

    def fit(
        self,
        train_set: Union[Dataset, DataLoader],
        valid_set: Optional[Union[Dataset, DataLoader]] = None,
        train_loader_kwargs: dict = None,
        valid_loader_kwargs: dict = None,
    ) -> None:
        """Fit the modules to the dataset. Main function of the Trainer class.

        Args:
            train_set (Union[Dataset, DataLoader]): dataset for training.
            valid_set (Optional[Union[Dataset, DataLoader]], optional): dataset for
                validation. If not provided, validation will not be performed.
                Defaults to None.
            train_loader_kwargs (dict, optional): optional kwargs for train dataloader.
                Defaults to None.
            valid_loader_kwargs (dict, optional): optional kwargs for valid dataloader.
                Defaults to None.
        """
        raise NotImplementedError

    def _on_fit_start(self):
        # Initialize optimizers
        self._init_optimizers()

        # set first epoch index
        self.start_epoch = 0

        # Load latest checkpoint to resume training if interrupted
        if self.checkpointer is not None:
            if self.checkpointer.exists_checkpoint():
                logger.info("loading last checkpoint")
                if self.checkpointer.load(mode="last"):
                    self.start_epoch = self.checkpointer.next_epoch
                    self.step = self.checkpointer.step
                    logger.info(
                        "resuming training from epoch "
                        f"{self.start_epoch} in step: {self.step}"
                    )

        if self.start_epoch == 0:
            self._save_initial_weights()

    def _save_initial_weights(self):
        if self.checkpointer is not None:
            self.checkpointer.save_initial_weights()

    def _init_optimizers(self):
        if self.opt_class is not None:
            self.optimizer = self.opt_class(self.modules.parameters())

    def _on_fit_end(self, epoch: Optional[int] = None):
        pass

    def _on_stage_start(self, stage: Stage, epoch: Optional[int] = None):
        pass

    def _on_stage_end(self, stage: Stage, loss=None, epoch: Optional[int] = None):
        pass

    def _on_train_interval(self, epoch: Optional[int] = None):
        pass

    def _save_intra_epoch_ckpt(self):
        raise NotImplementedError

    def _forward(self, batch: torch.Tensor, stage: Stage):
        raise NotImplementedError

    def _evaluate_batch(self, batch: torch.Tensor, stage: Stage):
        out = self._forward(batch, stage=stage)
        loss = self._compute_loss(out, batch, stage=stage)
        return loss.detach().cpu()

    def _compute_loss(
        self, predictions: torch.Tensor, batch: torch.Tensor, stage: Stage
    ):
        raise NotImplementedError

    def _fit_batch(self, batch: torch.Tensor):
        self.optimizer.zero_grad()
        outputs = self._forward(batch, Stage.TRAIN)
        loss = self._compute_loss(outputs, batch, Stage.TRAIN)
        loss.backward()
        self.optimizer.step()
        return loss.detach().cpu()

    def _update_average(self, loss: torch.Tensor, avg_loss: torch.Tensor):
        if torch.isfinite(loss):
            avg_loss -= avg_loss / self.step
            avg_loss += float(loss) / self.step
        return avg_loss



import logging
import tarfile
from pathlib import Path
from typing import Callable, List, Union
from urllib.parse import urlparse

import torch
from PIL import Image
from torch.utils.data import Dataset

from frarch.utils.data import download_url
from frarch.utils.exceptions import DatasetNotFoundError

logger = logging.getLogger(__name__)

urls = {
    "images": "https://www.robots.ox.ac.uk/~vgg/data/pets/data/images.tar.gz",
    "annotations": "https://www.robots.ox.ac.uk/~vgg/data/pets/data/annotations.tar.gz",
}


class OxfordPets(Dataset):
    """Oxford Pets dataset object.

    Data loader for the Oxford Pets dataset for pet recognition. The dataset can be
    obtained from https://www.robots.ox.ac.uk/~vgg/data/pets/data/images.tar.gz and
    their corresponding labels in
    https://www.robots.ox.ac.uk/~vgg/data/pets/data/annotations.tar.gz.

    Args:
        subset (str): "train" or "valid". Subset to load. Defaults to "train".
        transform (Callable): a callable object that takes an `PIL.Image` object and
            returns a modified `PIL.Image` object. Defaults to None, which won't apply
            any transformation.
        target_transform (Callable): a callable object that the label data and returns
            modified label data. Defaults to None, which won't apply any transformation.
        download (bool): True for downloading and storing the dataset data in the `root`
            directory if it's not present. Defaults to True.
        root (Union[str, Path]): root directory for the dataset.
            Defaults to `~/.cache/frarch/datasets/oxford_pets/`.

    References:
        - https://www.robots.ox.ac.uk/~vgg/data/pets/

    Examples:
        Simple usage of the dataset class::

            from frarch.datasets import Mit67
            from frarch.utils.data import create_dataloader
            from torchvision.transforms import ToTensor

            dataset = OxfordPets( "train", ToTensor, None, True, "./data/")
            dataloader = create_dataloader(dataset)
            for batch_idx, (batch, labels) in enumerate(dataloader):
                # process batch
    """

    def __init__(
        self,
        subset: str = "train",
        transform: Callable = None,
        target_transform: Callable = None,
        download: bool = True,
        root: Union[str, Path] = "~/.cache/frarch/datasets/oxford_pets/",
    ) -> None:
        if subset not in ["train", "valid"]:
            raise ValueError(f"set must be train or test not {subset}")

        self.root = Path(root).expanduser()
        self.images_root = self.root / "images"

        self.set = subset
        self.transform = transform
        self.target_transform = target_transform

        self.train_lst_path = self.root / "annotations" / "trainval.txt"
        self.valid_lst_path = self.root / "annotations" / "test.txt"

        if download and not self._detect_dataset():
            self._download_dataset()
            self._download_annotations()
        if not self._detect_dataset():
            raise DatasetNotFoundError(
                f"download flag not set and dataset not present in {self.root}"
            )

        self._load_set()

        logger.info(
            f"Loaded {self.set} Split: {len(self.images)} instances"
            f" in {len(self.classes)} classes"
        )

    def __getitem__(self, index: int) -> Union[torch.Tensor, int]:
        path, target = self.images[index]
        img = Image.open(path).convert("RGB")
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return img, target

    def __len__(self) -> int:
        return len(self.images)

    def get_number_classes(self) -> int:
        """Get number of target labels.

        Returns:
            int: number of target labels.
        """
        return len(self.classes)

    def _download_annotations(self) -> None:
        self._download_file("images")

    def _download_dataset(self) -> None:
        self._download_file("annotations")

    def _download_file(self, url_key: str) -> None:
        self.root.mkdir(parents=True, exist_ok=True)

        # download train/val images/annotations
        parts = urlparse(urls[url_key])
        filename = Path(parts.path).name
        cached_file = self.root / filename

        if not cached_file.exists():
            logger.info('Downloading: "{}" to {}\n'.format(urls[url_key], cached_file))
            download_url(urls[url_key], cached_file)

        # extract file
        logger.info(f"Extracting tar file {cached_file} to {self.root}")
        tar = tarfile.open(cached_file, "r")
        tar.extractall(self.root)
        tar.close()
        logger.info(f"Done! Removing dached file {cached_file}...")
        cached_file.unlink()

    def _get_file_paths(self) -> List[Path]:
        return list(self.images_root.glob("*.jpg"))

    def _detect_dataset(self) -> bool:
        if not self.root.exists():
            return False
        else:
            num_images = len(self._get_file_paths())
            annotations_present = (
                self.train_lst_path.exists() and self.valid_lst_path.exists()
            )
            return num_images > 0 and annotations_present

    def _load_set(self) -> None:
        path = self.train_lst_path if self.set == "train" else self.valid_lst_path
        with path.open("r") as f:
            self.images = []
            self.classes = set()
            for line in f:
                path, id_, species, breed_id = line.strip().split(" ")
                id_ = int(id_) - 1
                full_path = self.images_root / (path + ".jpg")
                self.images.append((full_path, id_))
                self.classes.add(id_)



import torch
import torch.nn as nn
import torch.nn.functional as F


class MNISTCNN(nn.Module):
    """Small CNN network for MNIST dataset.

    Args:
        input_channels (int): Number of input channels in the input tensor.
            Defaults to 1.
        embedding_size (int): Size of the output embedding for the feature extraction
            network. Defaults to 256.
    """

    def __init__(self, input_channels: int = 1, embedding_size: int = 256) -> None:
        super(MNISTCNN, self).__init__()
        self.conv1 = nn.Conv2d(input_channels, 32, kernel_size=10, padding="same")
        self.conv2 = nn.Conv2d(32, 32, kernel_size=10, padding="same")
        self.conv3 = nn.Conv2d(32, 64, kernel_size=10, padding="same")
        self.fc1 = nn.Linear(64 * 7 * 7, embedding_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Define the computation performed at every call.

        forward computation for MNISTCNN.

        Args:
            x (torch.Tensor): input to the model.

        Returns:
            torch.Tensor: output of the model.
        """
        x = F.relu(self.conv1(x))
        x = F.relu(F.max_pool2d(self.conv2(x), 2))
        x = F.dropout(x, p=0.5, training=self.training)
        x = F.relu(F.max_pool2d(self.conv3(x), 2))
        x = F.dropout(x, p=0.5, training=self.training)
        x = x.view(-1, 64 * 7 * 7)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        return x


class MNISTClassifier(nn.Module):
    """Classifier network for MNISTCNN.

    Args:
        embedding_size (int): embedding size from MNISTCNN network. Defaults to 256.
        classes (int): number of output classes for the classifier. Defaults to 10.
    """

    def __init__(self, embedding_size: int = 256, num_classes: int = 10) -> None:
        super(MNISTClassifier, self).__init__()
        self.fc2 = nn.Linear(embedding_size, num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Define the computation performed at every call.

        forward computation for MNISTClassifier.

        Args:
            x (torch.Tensor): input to the model.

        Returns:
            torch.Tensor: output of the model.
        """
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)



import torch


class BypassModel(torch.nn.Module):
    """Bypass mock model.

    Module that returns the same argument that is given.
    """

    def __init__(self) -> None:
        super(BypassModel, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute forward for bypass model.

        Args:
            x (torch.Tensor): input to the model.

        Returns:
            torch.Tensor: same tensor as the input.
        """
        return x



from .fashionCNN import FashionClassifier, FashionCNN
from .mitCNNs import MitCNN, MitCNNClassifier
from .mnistCNN import MNISTCNN, MNISTClassifier
from .vgg import (
    VGG,
    VGGClassifier,
    vgg11,
    vgg11_bn,
    vgg13,
    vgg13_bn,
    vgg16,
    vgg16_bn,
    vgg19,
    vgg19_bn,
    vggclassifier,
)



from . import cnn
from .cnn import (
    MNISTCNN,
    VGG,
    FashionCNN,
    VGGClassifier,
    vgg11,
    vgg11_bn,
    vgg13,
    vgg13_bn,
    vgg16,
    vgg16_bn,
    vgg19,
    vgg19_bn,
    vggclassifier,
)



import random
from typing import Callable, Iterable, Optional, Union

import PIL.Image as im
import PIL.ImageEnhance as ie
import torch
from PIL.Image import Image

NoneType = type(None)


class RandomFlip(object):
    """Randomly flips the given PIL.Image.

    Probability of 0.25 horizontal flip, 0.25 vertical flip, 0.5 no flip.

    Example:
        Simple usage of the class::

            image = PIL.Image.open("testimage.jpg")
            random_flip = RandomFlip()
            processed_image = random_flip(image)
    """

    def __call__(self, img: Image) -> Image:
        if not isinstance(img, Image):
            raise ValueError(f"img is {type(img)} not a PIL.Image object")

        dispatcher = {
            0: img,
            1: img,
            2: img.transpose(im.FLIP_LEFT_RIGHT),
            3: img.transpose(im.FLIP_TOP_BOTTOM),
        }

        return dispatcher[random.randint(0, 3)]


class RandomRotate(object):
    """Randomly rotate the given PIL.Image.

    Probability of 1/6 90°, 1/6 180°, 1/6 270°, 1/2 as is.

    Example:
        Simple usage of the class::

            image = PIL.Image.open("testimage.jpg")
            random_rotate = RandomRotate()
            processed_image = random_rotate(image)
    """

    def __call__(self, img: Image) -> Image:
        if not isinstance(img, Image):
            raise ValueError(f"img is {type(img)} not a PIL.Image object")

        dispatcher = {
            0: img,
            1: img,
            2: img,
            3: img.transpose(im.ROTATE_90),
            4: img.transpose(im.ROTATE_180),
            5: img.transpose(im.ROTATE_270),
        }

        return dispatcher[random.randint(0, 5)]


class PILColorBalance(object):
    """Randomly dim or enhance color of an image.

    Randomly dim or enhance color of an image given as an input. Given var value,
    enhance color from a 1-var factor to a 1+var factor.

    Args:
        var (float): float value to get random alpha value from [1-var, 1+var].

    Example:
        Simple usage of the class::

            image = PIL.Image.open("testimage.jpg")
            random_color_balance = PILColorBalance()
            processed_image = random_color_balance(image)
    """

    def __init__(self, var: float) -> None:
        if not isinstance(var, float):
            raise ValueError(f"{self.__class__.__name__}.var must be a float value")
        if var < 0 or var > 1:
            raise ValueError(
                f"{self.__class__.__name__}.var must be a float value between 0 and 1"
            )
        self.var = var

    def __call__(self, img: Image) -> Image:
        if not isinstance(img, Image):
            raise ValueError(f"img is {type(img)} not a PIL.Image object")

        alpha = random.uniform(1 - self.var, 1 + self.var)
        return ie.Color(img).enhance(alpha)


class PILContrast(object):
    """Randomly dim or enhance contrast of an image.

    Randomly dim or enhance contrast of an image given as an input. Given var value,
    enhance contrast from a 1-var factor to a 1+var factor.

    Args:
        var (float): float value to get random alpha value from [1-var, 1+var].

    Example:
        Simple usage of the class::

            image = PIL.Image.open("testimage.jpg")
            random_contrast = PILContrast()
            processed_image = random_contrast(image)
    """

    def __init__(self, var: float) -> None:
        if not isinstance(var, float):
            raise ValueError(f"{self.__class__.__name__}.var must be a float value")
        if var < 0 or var > 1:
            raise ValueError(
                f"{self.__class__.__name__}.var must be a float value between 0 and 1"
            )
        self.var = var

    def __call__(self, img: Image) -> Image:
        if not isinstance(img, Image):
            raise ValueError(f"img is {type(img)} not a PIL.Image object")

        alpha = random.uniform(1 - self.var, 1 + self.var)
        return ie.Contrast(img).enhance(alpha)


class PILBrightness(object):
    """Randomly dim or enhance brightness of an image.

    Randomly dim or enhance brightness of an image given as an input. Given var value,
    enhance brightness from a 1-var factor to a 1+var factor.

    Args:
        var (float): float value to get random alpha value from [1-var, 1+var].

    Example:
        Simple usage of the class::

            image = PIL.Image.open("testimage.jpg")
            random_brightness = PILBrightness()
            processed_image = random_brightness(image)
    """

    def __init__(self, var: float) -> None:
        if not isinstance(var, float):
            raise ValueError(f"{self.__class__.__name__}.var must be a float value")
        if var < 0 or var > 1:
            raise ValueError(
                f"{self.__class__.__name__}.var must be a float value between 0 and 1"
            )
        self.var = var

    def __call__(self, img: Image) -> Image:
        if not isinstance(img, Image):
            raise ValueError(f"img is {type(img)} not a PIL.Image object")

        alpha = random.uniform(1 - self.var, 1 + self.var)
        return ie.Brightness(img).enhance(alpha)


class PILSharpness(object):
    """Randomly dim or enhance sharpness of an image.

    Randomly dim or enhance sharpness of an image given as an input. Given var value,
    enhance sharpness from a 1-var factor to a 1+var factor.

    Args:
        var (float): float value to get random alpha value from [1-var, 1+var].

    Example:
        Simple usage of the class::

            image = PIL.Image.open("testimage.jpg")
            random_sharpness = PILSharpness()
            processed_image = random_sharpness(image)
    """

    def __init__(self, var: float) -> None:
        if not isinstance(var, float):
            raise ValueError(f"{self.__class__.__name__}.var must be a float value")
        if var < 0 or var > 1:
            raise ValueError(
                f"{self.__class__.__name__}.var must be a float value between 0 and 1"
            )
        self.var = var

    def __call__(self, img: Image) -> Image:
        if not isinstance(img, Image):
            raise ValueError(f"img is {type(img)} not a PIL.Image object")

        alpha = random.uniform(1 - self.var, 1 + self.var)
        return ie.Sharpness(img).enhance(alpha)


class RandomOrder(object):
    """Composes several transforms together in random order.

    Args:
        transforms (Iterable[Callable]): An iterable containing callable objects
            which take PIL.Image as input and outputs the same object type.

    Example:
        Simple usage of the class::

            image = PIL.Image.open("testimage.jpg")
            random_sharpness = RandomOrder(
                [
                    RandomFlip(),
                    RandomRotate(),
                    PILColorBalance(0.1),
                    PILContrast(0.1),
                    PILBrightness(0.1),
                    PILSharpness(0.1),
                ]
            )
            processed_image = random_sharpness(image)
    """

    def __init__(self, transforms: Optional[Iterable[Callable]]) -> None:
        if not isinstance(transforms, (Iterable, NoneType)):
            raise ValueError("transforms must be an iterable object")
        if transforms is not None:
            if len(transforms) == 0:
                transforms = None
            elif not all(isinstance(x, Callable) for x in transforms):
                raise ValueError("all objects in transforms must be callable")
        self.transforms = transforms

    def __call__(self, img: Image) -> Image:
        if not isinstance(img, Image):
            raise ValueError(f"img is {type(img)} not a PIL.Image object")

        if self.transforms is None:
            return img
        order = torch.randperm(len(self.transforms))
        for i in order:
            img = self.transforms[i](img)
        return img


class PowerPIL(RandomOrder):
    """Composes several transforms together in random order.

    Args:
        transforms (Iterable[Callable]): An iterable containing callable objects
            which take PIL.Image as input and outputs the same object type.

    Example:
        Simple usage of the class::

            image = PIL.Image.open("testimage.jpg")
            powerpil = PowerPIL(True, True, 0.1, 0.1, 0.1, 0.1)
            processed_image = powerpil(image)
    """

    def __init__(
        self,
        rotate: bool = True,
        flip: bool = True,
        colorbalance: float = 0.4,
        contrast: float = 0.4,
        brightness: float = 0.4,
        sharpness: float = 0.4,
    ) -> None:
        self._check_parameters(
            rotate, flip, colorbalance, contrast, brightness, sharpness
        )
        self.transforms = []
        if rotate:
            self.transforms.append(RandomRotate())
        if flip:
            self.transforms.append(RandomFlip())
        if brightness != 0:
            self.transforms.append(PILBrightness(brightness))
        if contrast != 0:
            self.transforms.append(PILContrast(contrast))
        if colorbalance != 0:
            self.transforms.append(PILColorBalance(colorbalance))
        if sharpness != 0:
            self.transforms.append(PILSharpness(sharpness))

    @staticmethod
    def _check_parameters(
        rotate, flip, colorbalance, contrast, brightness, sharpness
    ) -> None:
        if not isinstance(rotate, bool):
            raise ValueError("rotate must be boolean")
        if not isinstance(flip, bool):
            raise ValueError("flip must be boolean")

        if isinstance(colorbalance, float):
            if not 0 <= colorbalance <= 1:
                raise ValueError("colorbalance must be float between 0 and 1")
        else:
            raise ValueError("colorbalance must be float between 0 and 1")

        if isinstance(contrast, float):
            if not 0 <= contrast <= 1:
                raise ValueError("contrast must be float between 0 and 1")
        else:
            raise ValueError("contrast must be float between 0 and 1")

        if isinstance(brightness, float):
            if not 0 <= brightness <= 1:
                raise ValueError("brightness must be float between 0 and 1")
        else:
            raise ValueError("brightness must be float between 0 and 1")

        if isinstance(sharpness, float):
            if not 0 <= sharpness <= 1:
                raise ValueError("sharpness must be float between 0 and 1")
        else:
            raise ValueError("sharpness must be float between 0 and 1")



from pathlib import Path
from typing import Union


class DatasetNotFoundError(Exception):
    """Exception raised for OS dataset errors.

    Args:
        path ([Path, str]): path where the dataset should be
    """

    def __init__(self, path: Union[str, Path], msg="Dataset not found in path {path}"):
        self.path = path
        self.msg = msg
        super().__init__(self.msg.format(path=path))



from typing import Any, Dict

from .base import Metric


class MetricsWrapper:
    """Store a set of metrics and perform operations in all of them simultaneously.

    Example:
        Sample code for metrics wrapper:
            metrics_wrapper = MetricsWrapper(
                metric_str0=Metric0(),
                metric_str1=Metric1(),
            )
            model = Model()

            for batch, labels in dataset:
                predictions = model(batch)
                metrics_wrapper.update(predictions, labels)
            print(metrics_wrapper.get_metrics())
            # prints {"metric_str0": 0.0, "metric_str1": 1.0}
    """

    def __init__(self, **kwargs: Metric) -> None:
        """Initialize metrics in wrapper.

        Raises:
            ValueError: if any of the values in kwargs don't inherit from metric.
        """
        for k, v in kwargs.items():
            if not isinstance(v, Metric):
                raise ValueError(f"value for key {k} should inherit from Metric")
            setattr(self, k, v)

    def reset(self) -> None:
        """Call reset in all metrics in the MetricsWrapper class."""
        for _, v in self.__dict__.items():
            v.reset()

    def update(self, *args: Any, **kwargs: Any) -> None:
        """Call update on all metrics in MetricsWrapper class."""
        for _, v in self.__dict__.items():
            v.update(*args, **kwargs)

    def get_metrics(self, *args: Any, **kwargs: Any) -> Dict[str, Metric]:
        """Build a dict with aggregated metrics.

        Returns:
            Dict[str, Metric]: dict with metric names as keys and aggregated metrics\
                as values.
        """
        metrics = {}
        for k, v in self.__dict__.items():
            metrics[k] = v.get_metric(*args, **kwargs)
        return metrics



import abc
from typing import Any

import torch

AGGREGATION_MODES = ["mean", "max", "min"]


class Metric(metaclass=abc.ABCMeta):
    """abstract class for Metric objects.

    Example:
        Simple usage of the Metric class::
            class MyMetric(Metric):
                def _update(self, predictions, truth):
                    # compute some metric
                    return metric_value

            model = MyModel()
            mymetric = MyMetric()
            for batch, labels in dataset:
                predictions = model(batch)
                mymetric.update(predictions, labels)
            print(mymetric.get_metric(mode="mean"))

    """

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """Clear metrics from class."""
        self.metrics = []

    def update(self, predictions: torch.Tensor, truth: torch.Tensor) -> None:
        """Compute metric value and append to the metrics array.

        Args:
            predictions (torch.Tensor): output tensors from model.
            truth (torch.Tensor): ground truth tensor.
        """
        self.metrics.append(self._update(predictions, truth))

    @abc.abstractmethod
    def _update(self, predictions: torch.Tensor, truth: torch.Tensor) -> Any:
        """Compute the metric value.

        Args:
            predictions (torch.Tensor): output tensors from model.
            truth (torch.Tensor): ground truth tensor.
        """

    def __len__(self) -> int:
        return len(self.metrics)

    def get_metric(self, mode="mean") -> float:
        """Aggregate all values stored in the metric class.

        Args:
            mode (str, optional): aggregation type. mean, max or min.
                Defaults to "mean".

        Raises:
            ValueError: aggregation mode not supported

        Returns:
            float: aggregated metric.
        """
        if len(self) == 0:
            return 0.0

        if mode not in AGGREGATION_MODES:
            raise ValueError(
                f"Mode {mode} not supported. Supported modes: {AGGREGATION_MODES}"
            )
        if mode == "mean":
            return sum(self.metrics) / len(self)
        elif mode == "max":
            return max(self.metrics)
        elif mode == "min":
            return min(self.metrics)



import json
import logging
import random
import tarfile
from collections import Counter
from pathlib import Path
from typing import Callable, List, Union
from urllib.parse import urlparse

import torch
from PIL import Image
from torch.utils.data import Dataset

from frarch.utils.data import download_url
from frarch.utils.exceptions import DatasetNotFoundError

logger = logging.getLogger(__name__)

urls = {
    "images": "http://groups.csail.mit.edu/vision/LabelMe/NewImages/indoorCVPR_09.tar"
}


class Mit67(Dataset):
    """Mit 67 dataset object.

    Data loader for the Mit 67 dataset for indoor scene recognition. The dataset can
    be obtained from
    http://groups.csail.mit.edu/vision/LabelMe/NewImages/indoorCVPR_09.tar.

    Args:
        train (bool): True for loading the train subset and False for valid. Defaults
            to True.
        transform (Callable): a callable object that takes an `PIL.Image` object and
            returns a modified `PIL.Image` object. Defaults to None, which won't apply
            any transformation.
        target_transform (Callable): a callable object that the label data and returns
            modified label data. Defaults to None, which won't apply any transformation.
        download (bool): True for downloading and storing the dataset data in the `root`
            directory if it's not present. Defaults to True.
        root (Union[str, Path]): root directory for the dataset.
            Defaults to `~/.cache/frarch/datasets/mit67/`.

    References:
        - http://web.mit.edu/torralba/www/indoor.html

    Examples:
        Simple usage of the dataset class::

            from frarch.datasets import Mit67
            from frarch.utils.data import create_dataloader
            from torchvision.transforms import ToTensor

            dataset = Mit67(True, ToTensor, None, True, "./data/")
            dataloader = create_dataloader(dataset)
            for batch_idx, (batch, labels) in enumerate(dataloader):
                # process batch
    """

    def __init__(
        self,
        train: bool = True,
        transform: Callable = None,
        target_transform: Callable = None,
        download: bool = True,
        root: Union[str, Path] = "~/.cache/frarch/datasets/mit67/",
    ) -> None:
        self.root = Path(root).expanduser()
        self.set = "train" if train else "test"
        self.transform = transform
        self.target_transform = target_transform

        self.train_lst_path = self.root / "train.lst"
        self.valid_lst_path = self.root / "valid.lst"
        self.mapper_path = self.root / "class_map.json"

        if download and not self._detect_dataset():
            self._download_mit_dataset()
        if not self._detect_dataset():
            raise DatasetNotFoundError(
                f"download flag not set and dataset not present in {self.root}"
            )

        self._build_and_load_data_files()

        logger.info(
            f"Loaded {self.set} Split: {len(self.images)} instances"
            f" in {len(self.classes)} classes"
        )

    def __getitem__(self, index: int) -> Union[torch.Tensor, int]:
        path, target = self.images[index]
        img = Image.open(path).convert("RGB")
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return img, target

    def __len__(self) -> int:
        return len(self.images)

    def get_number_classes(self) -> int:
        """Get number of target labels.

        Returns:
            int: number of target labels.
        """
        return len(self.classes)

    def _download_mit_dataset(self) -> None:
        self.root.mkdir(parents=True, exist_ok=True)

        # download train/val images/annotations
        parts = urlparse(urls["images"])
        filename = Path(parts.path).name
        cached_file = self.root / filename

        if not cached_file.exists():
            logger.info('Downloading: "{}" to {}\n'.format(urls["images"], cached_file))
            download_url(urls["images"], cached_file)

        # extract file
        logger.info(f"[dataset] Extracting tar file {cached_file} to {self.root}")
        tar = tarfile.open(cached_file, "r")
        tar.extractall(self.root)
        tar.close()
        logger.info("[dataset] Done!")
        cached_file.unlink()

    def _get_file_paths(self) -> List[Path]:
        return list(self.root.glob("Images/*/*.jpg"))

    def _detect_dataset(self) -> bool:
        if not self.root.exists():
            return False
        else:
            num_images = len(self._get_file_paths())
            return num_images > 0

    def _build_and_load_data_files(self) -> None:
        all_paths = self._get_file_paths()
        self._load_class_map(all_paths)
        self._load_train_test_files(all_paths)

    def _load_class_map(self, all_paths: List[Path]) -> None:
        if not self.mapper_path.exists():
            self._build_class_mapper(all_paths)
        with self.mapper_path.open("r") as f:
            self.classes = json.load(f)

    def _build_class_mapper(self, all_paths: List[Path]) -> None:
        classes_set = set(map(lambda path: path.parts[-2], all_paths))
        logger.info(f"found {len(classes_set)} classes.")
        class_mapper = dict(zip(classes_set, range(len(classes_set))))
        logger.info(f"class mapper built: {class_mapper}")
        with self.mapper_path.open("w") as f:
            json.dump(class_mapper, f)

    def _load_train_test_files(self, all_paths: List[Path]) -> None:
        if not self.train_lst_path.exists() and not self.valid_lst_path.exists():
            self._build_train_test_files(all_paths)
        self._load_set(self.set)

    def _build_train_test_files(self, all_paths: List[Path]) -> None:
        classes_list = list(map(lambda path: path.parts[-2], all_paths))
        instance_counter = Counter(classes_list)

        train_instances, valid_instances = [], []
        for class_name, count in instance_counter.items():
            class_instances = list(
                filter(lambda x: x.parts[-2] == class_name, all_paths)
            )
            random.shuffle(class_instances)
            valid_count = max(1, int(count / 10))
            class_valid_instances = class_instances[:valid_count]
            class_train_instances = class_instances[valid_count:]

            valid_instances.extend(
                [
                    (instance, self.classes[class_name])
                    for instance in class_valid_instances
                ]
            )
            train_instances.extend(
                [
                    (instance, self.classes[class_name])
                    for instance in class_train_instances
                ]
            )

        logger.info(
            f"Built Train Split: {len(train_instances)} instances"
            f" in {len(self.classes)} classes"
        )
        logger.info(
            f"Built Valid Split: {len(valid_instances)} instances"
            f" in {len(self.classes)} classes"
        )

        with self.train_lst_path.open("w") as f:
            for line in train_instances:
                f.write(",".join(map(str, line)) + "\n")
        with self.valid_lst_path.open("w") as f:
            for line in valid_instances:
                f.write(",".join(map(str, line)) + "\n")

    def _load_set(self, set: str) -> None:
        path = self.train_lst_path if set == "train" else self.valid_lst_path
        with path.open("r") as f:
            self.images = []
            for line in f:
                path, label = line.strip().split(",")
                self.images.append((path, int(label)))



import logging
import sys
from pathlib import Path
from typing import Union


def create_logger_file(log_file_path: Union[str, Path], debug=False, stdout=False):
    """Create logger file in file path.

    Args:
        log_file_path (Union[str, Path]): path to logger file.
        debug (bool, optional): log debug statements. Defaults to False.
        stdout (bool, optional): log to stdout as well as to the log file.
            Defaults to False.

    Raises:
        ValueError: log_file_path is not a str or pathlib.Path object.
        ValueError: debug is not boolean.
        ValueError: stdout is not boolean.
    """
    if not isinstance(log_file_path, (str, Path)):
        raise ValueError("path must be a string or Path object")
    if not isinstance(debug, bool):
        raise ValueError("debug must be boolean")
    if not isinstance(stdout, bool):
        raise ValueError("stdout must be boolean")

    # cast to path
    log_file_path = Path(log_file_path)

    # get logging level
    level = logging.DEBUG if debug else logging.INFO

    # set logging format
    logging_format = (
        "[%(asctime)s][%(filename)s(%(lineno)d):%(funcName)s]"
        "-%(levelname)s: %(message)s"
    )

    # cretate parent and create file handler for config
    log_file_path.parent.mkdir(parents=True, exist_ok=True)
    logging.basicConfig(filename=log_file_path, level=level, format=logging_format)

    # if stdout add console_handler
    if stdout:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(logging.Formatter(logging_format))
        logging.getLogger().addHandler(console_handler)



"""
VGG definition. Slightly modified pytorch implementation.

:Description: VGG

:Authors: victor badenas (victor.badenas@gmail.com),
          pytorch.org

:Version: 0.1.0
:Created on: 21/07/2021 19:00
"""

import logging
from collections import OrderedDict
from typing import Any, Dict, Iterable, List, Mapping, Union, cast

import torch
import torch.nn as nn
import torchvision
from torch.hub import load_state_dict_from_url

__all__ = torchvision.models.vgg.__all__

l11 = [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"]
l13 = [64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"]
l16 = [
    64,
    64,
    "M",
    128,
    128,
    "M",
    256,
    256,
    256,
    "M",
    512,
    512,
    512,
    "M",
    512,
    512,
    512,
    "M",
]
l19 = [
    64,
    64,
    "M",
    128,
    128,
    "M",
    256,
    256,
    256,
    256,
    "M",
    512,
    512,
    512,
    512,
    "M",
    512,
    512,
    512,
    512,
    "M",
]

CONFIGURATIONS: Dict[str, List[Union[str, int]]] = {
    "11": l11,
    "13": l13,
    "16": l16,
    "19": l19,
}

model_urls = {
    "vgg11": "https://download.pytorch.org/models/vgg11-8a719046.pth",
    "vgg13": "https://download.pytorch.org/models/vgg13-19584684.pth",
    "vgg16": "https://download.pytorch.org/models/vgg16-397923af.pth",
    "vgg19": "https://download.pytorch.org/models/vgg19-dcbb9e9d.pth",
    "vgg11_bn": "https://download.pytorch.org/models/vgg11_bn-6002323d.pth",
    "vgg13_bn": "https://download.pytorch.org/models/vgg13_bn-abd245e5.pth",
    "vgg16_bn": "https://download.pytorch.org/models/vgg16_bn-6c64b313.pth",
    "vgg19_bn": "https://download.pytorch.org/models/vgg19_bn-c79401a0.pth",
}


class VGG(nn.Module):
    """VGG network definition.

    From `"Very Deep Convolutional Networks for Large-Scale Image Recognition"
    <https://arxiv.org/pdf/1409.1556.pdf>`_.

    Args:
        layers_cfg (List[Union[str, int]]): vgg layer configuration.
        batch_norm (bool, optional): Boolean flag for doing batch normalization.
            Defaults to True.
        init_weights (bool, optional): Force weight initialization.
            Defaults to True.
        pretrained (bool, optional): Get pretrained model. Defaults to False.
    """

    def __init__(
        self,
        layers_cfg: List[Union[str, int]],
        batch_norm: bool = True,
        init_weights: bool = True,
        pretrained: bool = False,
    ) -> None:
        super(VGG, self).__init__()
        self.layers_cfg = layers_cfg
        self.batch_norm = batch_norm
        self.features = self._make_layers(layers_cfg, batch_norm)
        self.avgpool = nn.AdaptiveAvgPool2d((7, 7))
        if init_weights:
            self._initialize_weights()
        elif pretrained:
            self._load_pretrained(layers_cfg, batch_norm)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Define the computation performed at every call.

        forward computation for VGG.

        Args:
            x (torch.Tensor): input to the model.

        Returns:
            torch.Tensor: output of the model.
        """
        x = self.features(x)
        x = self.avgpool(x)
        return torch.flatten(x, 1)

    def _initialize_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)

    def _make_layers(
        self, layers_cfg: List[Union[str, int]], batch_norm: bool = False
    ) -> nn.Sequential:
        layers: List[nn.Module] = []
        in_channels = 3
        for v in layers_cfg:
            if v == "M":
                layers.append(nn.MaxPool2d(kernel_size=2, stride=2))
            else:
                v = cast(int, v)
                conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding="same")
                layers.append(conv2d)
                if batch_norm:
                    layers.append(nn.BatchNorm2d(v))
                layers.append(nn.ReLU(inplace=True))
                in_channels = v
        return nn.Sequential(*layers)

    def _load_pretrained(self, layers_cfg, batch_norm):
        arch = "vgg"
        for model_depth, layers_preconfig in CONFIGURATIONS.items():
            if layers_cfg == layers_preconfig:
                arch += model_depth
                if batch_norm:
                    arch += "_bn"
                state_dict = load_state_dict_from_url(model_urls[arch], progress=True)
                features_state_dict, _ = split_state_dict(
                    state_dict, "features", "classifier"
                )
                self.load_state_dict(features_state_dict)
                return
        else:
            logging.warning(
                "no match found for pretrained models in the pytorch repository"
            )
            logging.warning("possible values:")
            for k, v in CONFIGURATIONS.items():
                logging.warning(f"vgg{k}: {v}")
            logging.warning("pretrained model not loaded. Starting from scratch.")


class VGGClassifier(nn.Module):
    """VGG classifier network.

    Args:
        num_classes ([type]): number of output classes.
        init_weights (bool, optional): Initialize weights. Defaults to True.
        pretrained (bool, optional): Pretrained architecture. Defaults to False.
        arch (str, optional): Architecture to load as pretrained. Defaults to "".
    """

    def __init__(
        self,
        num_classes: int,
        init_weights: bool = True,
        pretrained: bool = False,
        arch: str = "",
    ) -> None:
        super(VGGClassifier, self).__init__()
        self.in_features = 25088
        self.num_classes = num_classes
        self.classifier = nn.Sequential(
            nn.Linear(self.in_features, 4096),
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(4096, 4096),
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(4096, num_classes),
        )
        if init_weights:
            self._initialize_weights()
        elif pretrained and arch != "":
            self._load_pretrained(arch)

    def _initialize_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)

    def _load_pretrained(self, arch: str) -> None:
        state_dict = load_state_dict_from_url(model_urls[arch], progress=True)
        _, classifier_state_dict = split_state_dict(
            state_dict, "features", "classifier"
        )
        self.load_state_dict(classifier_state_dict)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Define the computation performed at every call.

        forward computation for VGG.

        Args:
            x (torch.Tensor): input to the model.

        Returns:
            torch.Tensor: output of the model.
        """
        x = self.classifier(x)
        return x


def split_state_dict(
    state_dict: Mapping, *search_strings: Iterable[str]
) -> List[Mapping]:
    results = [OrderedDict() for _ in range(len(search_strings))]
    for i, string in enumerate(search_strings):
        for k in state_dict.keys():
            if string in k:
                results[i][k] = state_dict.get(k)
    return results


def _vgg(cfg: str, batch_norm: bool, pretrained: bool, **kwargs: Any) -> VGG:
    if pretrained:
        kwargs["init_weights"] = False
    model = VGG(
        CONFIGURATIONS[cfg], batch_norm=batch_norm, pretrained=pretrained, **kwargs
    )
    return model


def vggclassifier(
    pretrained: bool, num_classes: int = 1000, arch: str = "", **kwargs: Any
) -> VGGClassifier:
    if pretrained and arch not in model_urls:
        logging.warning(
            "No architecture specified or arch does not exist, not loading model"
        )
        logging.warning(f"Pretrained model possibilities: {list(model_urls.keys())}")
        pretrained = False

    if pretrained and num_classes != 1000:
        logging.warning(
            "when loading pretrained classifiers, num_classes must"
            " be 1000, defaulting to random weights"
        )
        pretrained = False

    if pretrained:
        kwargs["init_weights"] = False

    model = VGGClassifier(num_classes, pretrained=pretrained, arch=arch, **kwargs)
    return model


def vgg11(pretrained: bool = False, **kwargs: Any) -> VGG:
    """Create VGG 11-layer model.

    VGG 11-layer model from `"Very Deep Convolutional Networks for Large-Scale Image
    Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_. The required minimum input
    size of the model is 32x32.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    return _vgg("11", False, pretrained, **kwargs)


def vgg11_bn(pretrained: bool = False, **kwargs: Any) -> VGG:
    """Create VGG 11-layer with batch normalization model.

    VGG 11-layer model with batch normalization from `"Very Deep Convolutional Networks
    for Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_.
    The required minimum input size of the model is 32x32.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    return _vgg("11", True, pretrained, **kwargs)


def vgg13(pretrained: bool = False, **kwargs: Any) -> VGG:
    """Create VGG 13-layer model.

    VGG 13-layer model from `"Very Deep Convolutional Networks for Large-Scale Image
    Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_. The required minimum input
    size of the model is 32x32.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    return _vgg("13", False, pretrained, **kwargs)


def vgg13_bn(pretrained: bool = False, **kwargs: Any) -> VGG:
    """Create VGG 13-layer with batch normalization model.

    VGG 13-layer model with batch normalization from `"Very Deep Convolutional Networks
    for Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_.
    The required minimum input size of the model is 32x32.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    return _vgg("13", True, pretrained, **kwargs)


def vgg16(pretrained: bool = False, **kwargs: Any) -> VGG:
    """Create VGG 16-layer model.

    VGG 16-layer model from `"Very Deep Convolutional Networks for Large-Scale Image
    Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_. The required minimum input
    size of the model is 32x32.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    return _vgg("16", False, pretrained, **kwargs)


def vgg16_bn(pretrained: bool = False, **kwargs: Any) -> VGG:
    """Create VGG 16-layer with batch normalization model.

    VGG 16-layer model with batch normalization from `"Very Deep Convolutional Networks
    for Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_.
    The required minimum input size of the model is 32x32.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    return _vgg("16", True, pretrained, **kwargs)


def vgg19(pretrained: bool = False, **kwargs: Any) -> VGG:
    """Create VGG 19-layer model.

    VGG 19-layer model from `"Very Deep Convolutional Networks for Large-Scale Image
    Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_. The required minimum input
    size of the model is 32x32.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    return _vgg("19", False, pretrained, **kwargs)


def vgg19_bn(pretrained: bool = False, **kwargs: Any) -> VGG:
    """Create VGG 19-layer with batch normalization model.

    VGG 19-layer model with batch normalization from `"Very Deep Convolutional Networks
    for Large-Scale Image Recognition" <https://arxiv.org/pdf/1409.1556.pdf>`_.
    The required minimum input size of the model is 32x32.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    return _vgg("19", True, pretrained, **kwargs)



import argparse
from pathlib import Path
from typing import Any, Dict, Tuple


def parse_arguments() -> Tuple[str, Dict[str, Any]]:
    """Parse arguments from command line.

    Returns:
        params_file (str): hyperparams file path.
        args (Dict[str, Any]]): arguments from argparse.Namespace.
    """
    parser = argparse.ArgumentParser(
        description="Run an experiment",
    )
    parser.add_argument(
        "param_file",
        type=str,
        help="A yaml-formatted file using the extended YAML syntax. "
        "defined by HyperPyYaml.",
    )

    parser.add_argument(
        "--debug",
        default=False,
        action="store_true",
        help="Run the experiment with only a few batches for all "
        "datasets, to ensure code runs without crashing.",
    )

    parser.add_argument(
        "--debug_batches",
        type=int,
        default=2,
        help="Number of batches to run in debug mode.",
    )

    parser.add_argument(
        "--device",
        type=str,
        default="cuda:0",
        help="The device to run the experiment on (e.g. 'cuda:0')",
    )

    parser.add_argument(
        "--noprogressbar",
        default=False,
        action="store_true",
        help="This flag disables the data loop progressbars.",
    )

    parser.add_argument(
        "--ckpt_interval_minutes",
        type=float,
        help="Amount of time between saving intra-epoch checkpoints "
        "in minutes. If non-positive, intra-epoch checkpoints are not saved.",
        default=15,
    )

    parser.add_argument(
        "--log_file",
        type=Path,
        help="Amount of time between saving intra-epoch checkpoints "
        "in minutes. If non-positive, intra-epoch checkpoints are not saved.",
        default=None,
    )

    args = parser.parse_args().__dict__
    param_file = args.pop("param_file")

    return param_file, args



import json
import logging
import shutil
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Mapping, Union

import torch

logger = logging.getLogger(__name__)

METRIC_MODES = ["min", "max"]
LOAD_MODES = ["last", "best"]


class Checkpointer:
    """Class for managing checkpoints.

    Args:
        save_path (Union[str, Path]): folder to store the checkpoint and training
            data to.
        modules (Mapping[str, torch.nn.Module]): dict-like structure with modules.
        save_best_only (bool, optional): If true, save only the best model according to
            some metric. If True, reference metric should be specified. If False, save
            all end of epoch checkpoints. Defaults to False.
        reference_metric (str, optional): Metric to use to determine the best model when
            save_best_only is True. Must be a string in the keys of the modules
            dict-like structure. Defaults to None.
        mode (str, optional): min if lower is better, max if higher is better.
            Examples: min for error and max for accuracy. Defaults to "min".

    Raises:
        ValueError: modules are not a dict or torch.nn.ModuleDict instance
        ValueError: modules in modules dict-like don't have string keys or
            torch.nn.Module values
            ValueError: path must be a string or Path object
        ValueError: metadata key is reserved for the metadata.json object.
        ValueError: metric mode is not min or max
        ValueError: save_best_only is True and no metric is defined.
    """

    def __init__(
        self,
        save_path: Union[str, Path],
        modules: Mapping[str, torch.nn.Module],
        save_best_only: bool = False,
        reference_metric: str = None,
        mode: str = "min",
    ) -> None:
        if not isinstance(modules, (dict, torch.nn.ModuleDict)):
            raise ValueError("modules must be a dict or torch.nn.ModuleDict instance")
        elif not all(
            isinstance(k, str) and isinstance(v, torch.nn.Module)
            for k, v in modules.items()
        ):
            raise ValueError("modules must have string keys and torch.nn.Module values")
        if not isinstance(save_path, (str, Path)):
            raise ValueError("path must be a string or Path object")

        if "metadata" in modules:
            raise ValueError("metadata in modules is reserved for metadata json object")
        if mode not in METRIC_MODES:
            raise ValueError(f"metric mode must be in {METRIC_MODES} not {mode}")
        if save_best_only and reference_metric is None:
            raise ValueError(
                "specify the metric name to consider while save_best_only=True"
            )

        self.base_path = Path(save_path) / "save"
        self.base_path.mkdir(exist_ok=True, parents=True)
        self.modules = modules
        self.metadata = {}
        self.save_best_only = save_best_only
        self.best_metric = None
        self.reference_metric = reference_metric
        self.mode = mode

    def save_initial_weights(self) -> None:
        """Save weights with which the model has been initialized."""
        time_str = str(datetime.now())
        ckpt_folder = "initial_weights"
        paths = self._build_paths(ckpt_folder)
        self._save_modules(paths)
        self._save_metadata(
            time_str,
            paths["metadata"],
            epoch=-1,
            intra_epoch=False,
            step=0,
        )

    def save(
        self,
        epoch: int,
        current_step: int,
        intra_epoch: bool = False,
        extra_data: Dict = None,
        **metrics: Dict[str, Any],
    ) -> None:
        """Save checkpoint.

        Args:
            epoch (int): current epoch index.
            current_step (int): current batch index.
            intra_epoch (bool, optional): boolean flag to indicate if the checkpoint is
                intra epoch if true and end of epoch if false. Defaults to False.
            extra_data (Dict, optional): extra metadata in json format to add to the
                metadata.json file. Defaults to None.
        """
        time_str = str(datetime.now())
        ckpt_folder = f"ckpt_{time_str.replace(' ', '_')}"

        paths = self._build_paths(ckpt_folder)
        self._save_modules(paths)
        self._save_metadata(
            time_str,
            paths["metadata"],
            epoch=epoch,
            intra_epoch=intra_epoch,
            step=current_step,
            extra_data=extra_data,
            **metrics,
        )
        if intra_epoch:
            logger.info(f"Saved intra_epoch model to {ckpt_folder}")
        else:
            logger.info(f"Saved end_of_epoch model to {ckpt_folder}")

        if not intra_epoch and self.reference_metric is not None:
            self._update_best_metric(**metrics)
            if self.save_best_only:
                self._remove_old_ckpts(ckpt_folder)

    def load(self, mode="last", **load_kwargs) -> None:
        """Load checkpoint from folder.

        Args:
            mode (str, optional): last for loading the last checkpoint stored and best
                to load the model with the mest metric. Defaults to "last".

        Raises:
            ValueError: mode is not best or last
        """
        if mode == "best":
            return self._load_best_checkpoint(**load_kwargs)
        elif mode == "last":
            return self._load_last_checkpoint(**load_kwargs)
        else:
            raise ValueError('load\'s mode kwarg can be "best" or "last"')

    def exists_checkpoint(self) -> bool:
        """Check if save_path contains a checkpoint folder.

        Returns:
            bool: True if it contains a checkpoint folder, False if not.
        """
        for folder in self.base_path.iterdir():
            if self._is_ckpt_dir(folder):
                return True
        return False

    @property
    def current_epoch(self) -> int:
        if len(self.metadata) >= 0 and "epoch" in self.metadata:
            return int(self.metadata["epoch"])
        return 0

    @property
    def next_epoch(self) -> int:
        if self._is_intraepoch():
            return self.current_epoch
        else:
            return self.current_epoch + 1

    @property
    def step(self) -> int:
        if self._is_intraepoch():
            return self.metadata["step"]
        return 0

    def _build_paths(self, ckpt_folder_name: str) -> dict:
        paths = {}
        for module_name in self.modules.keys():
            module_path = self.base_path / ckpt_folder_name / f"{module_name}.pt"
            paths[module_name] = module_path
        paths["metadata"] = self.base_path / ckpt_folder_name / "metadata.json"
        return paths

    def _save_modules(self, paths: Dict[str, Path]):
        for module_name in self.modules:
            paths[module_name].parent.mkdir(exist_ok=True, parents=True)
            torch.save(
                self.modules[module_name].state_dict(),
                paths[module_name],
            )

    def _save_metadata(
        self,
        time_str: str,
        metadata_path: Union[Path, str],
        epoch: int,
        step: int,
        intra_epoch: bool = False,
        extra_data: Dict = None,
        **metrics,
    ) -> None:
        self.metadata = {
            "intra_epoch": intra_epoch,
            "step": step,
            "epoch": epoch,
            "time": time_str,
            "extra_info": extra_data,
            **metrics,
        }

        with open(metadata_path, "w") as metadata_handler:
            json.dump(self.metadata, metadata_handler, indent=4)

    def _update_best_metric(self, **metrics):
        if self.best_metric is None:
            self.best_metric = metrics[self.reference_metric]
        else:
            if self._is_better(metrics[self.reference_metric], self.best_metric):
                self.best_metric = metrics[self.reference_metric]

    def _remove_old_ckpts(self, curr_ckpt_folder):
        for old_ckpt in self.base_path.iterdir():
            if old_ckpt.name == curr_ckpt_folder or not str(old_ckpt.name).startswith(
                "ckpt_"
            ):
                continue
            with open(old_ckpt / "metadata.json", "r") as metadata_handler:
                old_metadata = json.load(metadata_handler)

            if self.reference_metric not in old_metadata:
                shutil.rmtree(old_ckpt)
            elif not self._is_better(
                old_metadata[self.reference_metric], self.best_metric
            ):
                shutil.rmtree(old_ckpt)

    def _is_better(self, new_metric, old_metric) -> bool:
        if self.mode == "min":
            return new_metric <= old_metric
        elif self.mode == "max":
            return new_metric >= old_metric

    def _load_best_checkpoint(self, **load_kwargs) -> None:
        ckpts_meta = self._load_checkpoints_meta()
        ckpts_meta.pop("initial_weights")
        cmp_fn = min if self.mode == "min" else max
        best_ckpt_name = cmp_fn(
            ckpts_meta, key=lambda i: ckpts_meta[i][self.reference_metric]
        )
        return self._load_checkpoint_from_folder(
            best_ckpt_name, ckpts_meta[best_ckpt_name], **load_kwargs
        )

    def _load_last_checkpoint(self, **load_kwargs) -> None:
        ckpts_meta = self._load_checkpoints_meta()
        latest_ckpt_name = max(ckpts_meta, key=lambda i: ckpts_meta[i]["time"])
        return self._load_checkpoint_from_folder(
            latest_ckpt_name, ckpts_meta[latest_ckpt_name], **load_kwargs
        )

    def _load_checkpoints_meta(self) -> dict:
        ckpts_meta = {}
        for folder in self.base_path.iterdir():
            if not folder.is_dir():
                continue

            if self._is_ckpt_dir(folder):
                metadata_path = folder / "metadata.json"
                with open(metadata_path, "r") as f:
                    ckpts_meta[folder.name] = json.load(f)

                ckpts_meta[folder.name]["time"] = datetime.strptime(
                    ckpts_meta[folder.name]["time"], "%Y-%m-%d %H:%M:%S.%f"
                )

        return ckpts_meta

    def _load_checkpoint_from_folder(
        self, ckpt_folder_name, metadata, **load_kwargs
    ) -> None:
        paths = self._build_paths(ckpt_folder_name)
        for module_name in self.modules:
            try:
                self.modules[module_name].load_state_dict(
                    torch.load(paths[module_name], **load_kwargs)
                )
            except Exception as e:
                logger.error(f"Failed loading ckpt from {ckpt_folder_name}.")
                logger.error(e)
                raise e
        self.metadata = metadata
        logger.info(
            f"Loaded ckpt from epoch {self.current_epoch} from {ckpt_folder_name}"
        )

    @staticmethod
    def _is_ckpt_dir(path: Union[str, Path]):
        return str(path.name).startswith("ckpt_") or str(path.name) == "initial_weights"

    def _is_intraepoch(self) -> bool:
        return self.metadata["intra_epoch"]



import json
import logging
import random
from collections import Counter
from pathlib import Path
from typing import Callable, Iterable, List, Mapping, Tuple, Union

import torch
from PIL import Image
from torch.utils.data import Dataset

from frarch.utils.exceptions import DatasetNotFoundError

logger = logging.getLogger(__name__)


class Caltech101(Dataset):
    """Caltech 101 dataset object.

    Data loader for the Caltech 101 dataset for object classification. The dataset can
    be obtained from
    http://www.vision.caltech.edu/Image_Datasets/Caltech101/101_ObjectCategories.tar.gz.

    Args:
        subset (str): "train" or "valid". Subset to load. Defaults to "train".
        transform (Callable): a callable object that takes an `PIL.Image` object and
            returns a modified `PIL.Image` object. Defaults to None, which won't apply
            any transformation.
        target_transform (Callable): a callable object that the label data and returns
            modified label data. Defaults to None, which won't apply any transformation.
        root (Union[str, Path]): root directory for the dataset. Defaults to `./data/`.

    References:
        - http://www.vision.caltech.edu/Image_Datasets/Caltech101/

    Examples:
        Simple usage of the dataset class::

            from frarch.datasets import Caltech101
            from frarch.utils.data import create_dataloader
            from torchvision.transforms import ToTensor
            dataset = Caltech101("train", ToTensor, None, "./data/")
            dataloader = create_dataloader(dataset)
            for batch_idx, (batch, labels) in enumerate(dataloader):
                # process batch
    """

    def __init__(
        self,
        subset: str = "train",
        transform: Callable = None,
        target_transform: Callable = None,
        root: Union[str, Path] = "./data/",
    ) -> None:
        if subset not in ["train", "valid"]:
            raise ValueError(f"set must be train or test not {subset}")

        self.root = Path(root).expanduser()

        self.set = subset
        self.transform = transform
        self.target_transform = target_transform

        self.train_lst_path = self.root / "train.lst"
        self.valid_lst_path = self.root / "valid.lst"
        self.mapper_path = self.root / "classes.json"

        if not self._detect_dataset():
            raise DatasetNotFoundError(
                self.root,
                "Dataset not found at {path}. Please download"
                " it from http://www.vision.caltech.edu/Image_Datasets/Caltech101/ .",
            )

        self._build_and_load_lst()

        logger.info(
            f"Loaded {self.set} Split: {len(self.images)} instances"
            f" in {len(self.classes)} classes"
        )

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int]:
        path, target = self.images[index]
        img = Image.open(path).convert("RGB")
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return img, target

    def __len__(self) -> int:
        return len(self.images)

    def get_number_classes(self) -> int:
        """Get number of target labels.

        Returns:
            int: number of target labels.
        """
        return len(self.classes)

    def _detect_dataset(self) -> bool:
        if not self.root.exists():
            return False
        else:
            num_images = len(self._get_file_paths())
            return num_images > 0

    def _build_and_load_lst(self) -> None:
        all_paths = self._get_file_paths()
        self._build_and_load_class_map(all_paths)
        self._load_train_test_files(all_paths)

    def _build_and_load_class_map(self, all_paths: List[Path]) -> None:
        if not self.mapper_path.exists():
            self._build_class_mapper(all_paths)
        self._load_class_map()

    def _build_class_mapper(self, all_paths: List[Path]) -> None:
        classes_set = set(map(lambda path: path.parts[-2], all_paths))
        logger.info(f"found {len(classes_set)} classes.")
        class_mapper = dict(zip(classes_set, range(len(classes_set))))
        logger.info(f"class mapper built: {class_mapper}")
        self._dump_class_map(class_mapper)

    def _load_train_test_files(self, all_paths: List[Path]) -> None:
        if not self.train_lst_path.exists() and not self.valid_lst_path.exists():
            self._build_train_test_files(all_paths)
        self._load_set()

    def _build_train_test_files(self, all_paths: List[Path]) -> None:
        classes_list = list(map(lambda path: path.parts[-2], all_paths))
        instance_counter = Counter(classes_list)

        train_instances, valid_instances = [], []
        for class_name, count in instance_counter.items():
            class_instances = list(
                filter(lambda x: x.parts[-2] == class_name, all_paths)
            )
            random.shuffle(class_instances)
            valid_count = max(1, int(count / 10))
            class_valid_instances = class_instances[:valid_count]
            class_train_instances = class_instances[valid_count:]

            valid_instances.extend(
                [
                    (instance, self.classes[class_name])
                    for instance in class_valid_instances
                ]
            )
            train_instances.extend(
                [
                    (instance, self.classes[class_name])
                    for instance in class_train_instances
                ]
            )

        logger.info(
            f"Built Train Split: {len(train_instances)} instances"
            f" in {len(self.classes)} classes"
        )
        logger.info(
            f"Built Valid Split: {len(valid_instances)} instances"
            f" in {len(self.classes)} classes"
        )

        self._write_lst(self.train_lst_path, train_instances)
        self._write_lst(self.valid_lst_path, valid_instances)

    @staticmethod
    def _write_lst(path: Path, instances: Iterable) -> None:
        with path.open("w") as f:
            for line in instances:
                f.write(",".join(map(str, line)) + "\n")

    def _get_file_paths(self) -> List[Path]:
        all_files = list(self.root.glob("*/*.jpg"))
        return list(filter(lambda x: x.parts[-2] != "BACKGROUND_Google", all_files))

    def _load_set(self) -> None:
        path = self.train_lst_path if self.set == "train" else self.valid_lst_path
        with path.open("r") as f:
            self.images = []
            for line in f:
                path, label = line.strip().split(",")
                self.images.append((path, int(label)))

    def _dump_class_map(self, class_mapper: Mapping) -> None:
        with self.mapper_path.open("w") as f:
            json.dump(class_mapper, f)

    def _load_class_map(self) -> None:
        with self.mapper_path.open("r") as f:
            self.classes = json.load(f)



import logging
from pathlib import Path
from typing import Any, Mapping, Optional, Union
from urllib.request import urlretrieve

import torch
from hyperpyyaml import resolve_references
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm

from frarch.utils.logging import create_logger_file

logger = logging.getLogger(__name__)


def create_dataloader(dataset: Dataset, **dataloader_kwargs) -> DataLoader:
    """Create dataloader from dataset.

    Args:
        dataset (torch.utils.data.Dataset): dataset object to feed onto DataLoader.

    Raises:
        ValueError: dataset is not a Dataset or does not inherit from it.

    Returns:
        torch.utils.data.DataLoader: dataloader with the dataset object.
    """
    if not isinstance(dataset, Dataset):
        raise ValueError("dataset needs to be a child or torch.utils.data.Dataset")
    return DataLoader(dataset, **dataloader_kwargs)


def build_experiment_structure(
    hparams_file: Union[str, Path],
    experiment_folder: Union[str, Path],
    overrides: Mapping = None,
    debug: bool = False,
):
    """Construct experiment folder hierarchy on experiment_folder.

    Args:
        hparams_file (Union[str, Path]): hparams configuration file path.
        experiment_folder (Union[str, Path]): Folder where to store experiment files.
            Defaults to "results/debug/".
        overrides (Mapping, optional): Parameters to override on hparams file.
            Defaults to None.
        debug (bool, optional): debug flag. Defaults to False.
    """
    if overrides is None:
        overrides = {}

    base_path = Path(experiment_folder)
    base_path.mkdir(exist_ok=True, parents=True)
    save_path = base_path / "save"
    save_path.mkdir(exist_ok=True, parents=True)

    experiment_yaml_path = base_path / "train.yaml"
    with open(hparams_file, "r") as f:
        hparams = resolve_references(f, overrides).getvalue()
    with open(experiment_yaml_path, "w") as f:
        f.write(hparams)

    log_file_path = base_path / "train.log"
    create_logger_file(log_file_path, debug=debug, stdout=debug)

    logger.info(f"experiment folder {str(base_path)} created successfully")


def download_url(
    url: str, destination: Optional[str] = None, progress_bar: bool = True
) -> str:
    """Download a URL to a local file.

    Args:
        url (str): The URL to download.
        destination (Optional[str], optional): The destination of the file. If None is
            given the file is saved to a temporary directory. Defaults to None.
        progress_bar (bool, optional): The destination of the file. If None is given
            the file is saved to a temporary directory. Defaults to True.

    Returns:
        str: filename of downloaded file.
    """

    def update_progressbar(t):
        last_b = [0]

        def inner(b=1, bsize=1, tsize=None):
            if tsize is not None:
                t.total = tsize
            if b > 0:
                t.update((b - last_b[0]) * bsize)
            last_b[0] = b

        return inner

    if progress_bar:
        with tqdm(unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]) as t:
            filename, _ = urlretrieve(
                url, filename=destination, reporthook=update_progressbar(t)
            )
    else:
        filename, _ = urlretrieve(url, filename=destination)
    return filename


def tensor_in_device(data: Any, device: str = "cpu", **kwargs) -> torch.Tensor:
    """Create tensor in device.

    Args:
        data (Any): data on the tensor.
        device (str, optional): string of the device to be created in.
            Defaults to "cpu".

    Returns:
        torch.Tensor: tensor in device.
    """
    return torch.Tensor(data, **kwargs).to(device)


def read_file(filepath: Union[str, Path]) -> str:
    """Read contents of file.

    Args:
        filepath (Union[str, Path]): path to file.

    Returns:
        str: contents of the file
    """
    return Path(filepath).read_text()



#!/usr/bin/env python
import glob
import os

from setuptools import find_packages, setup


def read(fname):
    """
    Read the contents of a file.

    Parameters
    ----------
    fname : str
        Path to file.

    Returns
    -------
    str
        File contents.
    """
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


install_requires = read("requirements.txt").splitlines()

extras_require = {}
extra_req_files = glob.glob("requirements-*.txt")
for extra_req_file in extra_req_files:
    name = os.path.splitext(extra_req_file)[0].replace("requirements-", "", 1)
    extras_require[name] = read(extra_req_file).splitlines()

if extras_require:
    extras_require["all"] = sorted({x for v in extras_require.values() for x in v})


meta = {}
exec(read("frarch/__meta__.py"), meta)

long_description = read("README.md")


setup(
    # Essential details on the package and its dependencies
    name=meta["name"],
    version=meta["version"],
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    package_dir={meta["name"]: os.path.join(".", meta["path"])},
    # If any package contains *.txt or *.rst files, include them:
    # package_data={"": ["*.txt", "*.rst"],}
    install_requires=install_requires,
    extras_require=extras_require,
    # Metadata to display on PyPI
    author=meta["author"],
    author_email=meta["author_email"],
    description=meta["description"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=meta["license"],
    url=meta["url"],
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import torch

from .base import Metric


class ClassificationError(Metric):
    """Classification error metric.

    Example:
        Sample code for use of the ClassificationError metric class:
            model = MyModel()
            error = ClassificationError()
            for batch, labels in dataset:
                predictions = model(batch)
                error.update(predictions, labels)
            print(error.get_metric(mode="mean"))
    """

    def _update(self, predictions: torch.Tensor, truth: torch.Tensor) -> torch.Tensor:
        if predictions.shape[0] != truth.shape[0]:
            raise ValueError(f"mismatched shapes {predictions.shape} != {truth.shape}")
        predictions = torch.argmax(predictions, dim=-1)
        return (predictions != truth).float().mean().item()



"""
ResNet definition. Slightly tweaked pytorch implementation.

:Description: ResNet

:Authors: victor badenas (victor.badenas@gmail.com),
          pytorch.org

:Version: 0.1.0
:Created on: 21/07/2021 19:00
"""

from typing import Any, Callable, List, Optional

import torch
import torch.nn as nn
import torchvision
from torch import Tensor
from torch.hub import load_state_dict_from_url

model_urls = torchvision.models.resnet.model_urls
__all__ = torchvision.models.resnet.__all__


def conv3x3(
    in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1
) -> nn.Conv2d:
    """3x3 convolution with padding."""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """1x1 convolution."""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class ResNetBlock(nn.Module):
    pass


class BasicBlock(ResNetBlock):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when
        # stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(ResNetBlock):
    # Bottleneck in torchvision places the stride for downsampling at 3x3
    # convolution(self.conv2) while original implementation places the stride at the
    # first 1x1 convolution(self.conv1) according to "Deep residual learning for image
    # recognition"https://arxiv.org/abs/1512.03385. This variant is also known as
    # ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample
        # the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    """Residual network architecture model.

    Args:
        block (ResNetBlock): basic construction block for architecture.
        layers (List[int]): layer configuration. I.e. [2, 2, 2, 2].
        num_classes (int)): output classes. Defaults to 1000.
        zero_init_residual (bool): True to initialize the residual layers with zeros.
            Default False.
        groups (int): groups for convolutional layers in residual blocks.
            Defaults to 1.
        width_per_group (int): Width for convolutional residual blocks. Defaults to 64.
        replace_stride_with_dilation (Optional[List[bool]]): Optional list of boolean
            values to replace stride with dilation. Defaults to None.
        norm_layer (Optional[Callable[..., nn.Module]]): Norm layer. If None, defaults
            to BatchNorm2d. Defautls to None.
    """

    def __init__(
        self,
        block: ResNetBlock,
        layers: List[int],
        num_classes: int = 1000,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Optional[List[bool]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super(ResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                "or a 3-element tuple, got {}".format(replace_stride_with_dilation)
            )
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv2d(
            3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False
        )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(
            block, 128, layers[1], stride=2, dilate=replace_stride_with_dilation[0]
        )
        self.layer3 = self._make_layer(
            block, 256, layers[2], stride=2, dilate=replace_stride_with_dilation[1]
        )
        self.layer4 = self._make_layer(
            block, 512, layers[3], stride=2, dilate=replace_stride_with_dilation[2]
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual
        # block behaves like an identity. This improves the model
        # by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(
        self,
        block: ResNetBlock,
        planes: int,
        blocks: int,
        stride: int = 1,
        dilate: bool = False,
    ) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def _forward_impl(self, x: Tensor) -> Tensor:
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

    def forward(self, x: Tensor) -> Tensor:
        """Define the computation performed at every call.

        forward computation for ResNet.

        Args:
            x (torch.Tensor): input to the model.

        Returns:
            torch.Tensor: output of the model.
        """
        return self._forward_impl(x)


def _resnet(
    arch: str,
    block: ResNetBlock,
    layers: List[int],
    pretrained: bool,
    progress: bool,
    **kwargs: Any,
) -> ResNet:
    model = ResNet(block, layers, **kwargs)
    if pretrained:
        state_dict = load_state_dict_from_url(model_urls[arch], progress=progress)
        model.load_state_dict(state_dict)
    return model


def resnet18(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> ResNet:
    """ResNet-18 model.

    From `"Deep Residual Learning for Image Recognition"
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet("resnet18", BasicBlock, [2, 2, 2, 2], pretrained, progress, **kwargs)


def resnet34(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> ResNet:
    """ResNet-34 model.

    From `"Deep Residual Learning for Image Recognition"
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet("resnet34", BasicBlock, [3, 4, 6, 3], pretrained, progress, **kwargs)


def resnet50(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> ResNet:
    """ResNet-50 model.

    From `"Deep Residual Learning for Image Recognition"
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet("resnet50", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs)


def resnet101(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> ResNet:
    """ResNet-101 model.

    From `"Deep Residual Learning for Image Recognition"
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet(
        "resnet101", Bottleneck, [3, 4, 23, 3], pretrained, progress, **kwargs
    )


def resnet152(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> ResNet:
    """ResNet-152 model.

    From `"Deep Residual Learning for Image Recognition"
    <https://arxiv.org/pdf/1512.03385.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet(
        "resnet152", Bottleneck, [3, 8, 36, 3], pretrained, progress, **kwargs
    )


def resnext50_32x4d(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> ResNet:
    """ResNeXt-50 32x4d model.

    From `"Aggregated Residual Transformation for Deep Neural Networks"
    <https://arxiv.org/pdf/1611.05431.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 4
    return _resnet(
        "resnext50_32x4d", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs
    )


def resnext101_32x8d(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> ResNet:
    """ResNeXt-101 32x8d model.

    From `"Aggregated Residual Transformation for Deep Neural Networks"
    <https://arxiv.org/pdf/1611.05431.pdf>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 8
    return _resnet(
        "resnext101_32x8d", Bottleneck, [3, 4, 23, 3], pretrained, progress, **kwargs
    )


def wide_resnet50_2(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> ResNet:
    """Wide ResNet-50-2 model.

    From `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_.
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["width_per_group"] = 64 * 2
    return _resnet(
        "wide_resnet50_2", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs
    )


def wide_resnet101_2(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> ResNet:
    """Wide ResNet-101-2 model.

    From `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_.
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs["width_per_group"] = 64 * 2
    return _resnet(
        "wide_resnet101_2", Bottleneck, [3, 4, 23, 3], pretrained, progress, **kwargs
    )



import torch
import torch.nn as nn


class FashionCNN(nn.Module):
    """Small CNN network for FashionMNIST dataset.

    Args:
        out_size (int): Size of the output embedding for the feature extraction
            network. Defaults to 128.
    """

    def __init__(self, out_size: int = 128) -> None:
        super(FashionCNN, self).__init__()

        self.layer1 = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2),
        )

        self.layer2 = nn.Sequential(
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.MaxPool2d(2),
        )

        self.fc1 = nn.Linear(in_features=64 * 7 * 7, out_features=512)
        self.drop = nn.Dropout2d(0.25)
        self.fc2 = nn.Linear(in_features=512, out_features=out_size)
        self.relu = nn.ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Define the computation performed at every call.

        forward computation for FashionCNN.

        Args:
            x (torch.Tensor): input to the model.

        Returns:
            torch.Tensor: output of the model.
        """
        out = self.layer1(x)
        out = self.layer2(out)
        out = out.view(out.size(0), -1)
        out = self.relu(self.fc1(out))
        out = self.drop(out)
        out = self.fc2(out)

        return out


class FashionClassifier(nn.Module):
    """Classifier network for FashionCNN.

    Args:
        embedding_size (int): embedding size from FashionCNN network. Defaults to 128.
        classes (int): number of output classes for the classifier. Defaults to 10.
    """

    def __init__(self, embedding_size: int = 128, classes: int = 10) -> None:
        super(FashionClassifier, self).__init__()
        self.fc = nn.Linear(in_features=embedding_size, out_features=classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Define the computation performed at every call.

        forward computation for FashionCNN.

        Args:
            x (torch.Tensor): input to the model.

        Returns:
            torch.Tensor: output of the model.
        """
        return self.fc(x)



"""
Base class for classifier network training.

:Description: base trainer class for training a classifier network

:Authors: victor badenas (victor.badenas@gmail.com)

:Version: 0.1.0
"""

__title__ = "classifier_trainer"
__version__ = "0.1.0"
__author__ = "victor badenas"

import logging
import time
from typing import Iterable, Optional, Union

import torch
from torch.utils.data import DataLoader, Dataset

from frarch.utils.data import create_dataloader
from frarch.utils.stages import Stage

from .base_trainer import BaseTrainer

logger = logging.getLogger(__name__)


class ClassifierTrainer(BaseTrainer):
    """Trainer class for classifiers.

    Args:
        modules (Mapping[str, torch.nn.Module]): trainable modules in the training.
        opt_class (Type[torch.optim.Optimizer]): optimizer class for training.
        hparams (Mapping[str, Any]): hparams dict-like structure from hparams file.
        checkpointer (Optional[Checkpointer], optional): Checkpointer class for saving
            the model and the hyperparameters needed. If None, no checkpoints are saved.
            Defaults to None.

    Raises:
        ValueError: ckpt_interval_minutes must be > 0 or None
        SystemError: Python version not supported. Python version must be >= 3.7
    """

    def fit(
        self,
        train_set: Union[Dataset, DataLoader],
        valid_set: Optional[Union[Dataset, DataLoader]] = None,
        train_loader_kwargs: dict = None,
        valid_loader_kwargs: dict = None,
    ) -> None:
        """Fit the modules to the dataset. Main function of the Trainer class.

        Args:
            train_set (Union[Dataset, DataLoader]): dataset for training.
            valid_set (Optional[Union[Dataset, DataLoader]], optional): dataset for
                validation. If not provided, validation will not be performed.
                Defaults to None.
            train_loader_kwargs (dict, optional): optional kwargs for train dataloader.
                Defaults to None.
            valid_loader_kwargs (dict, optional): optional kwargs for valid dataloader.
                Defaults to None.
        """
        if train_loader_kwargs is None:
            train_loader_kwargs = {}
        if valid_loader_kwargs is None:
            valid_loader_kwargs = {}

        if not isinstance(train_set, torch.utils.data.DataLoader):
            train_set = create_dataloader(train_set, **train_loader_kwargs)
        if valid_set is not None and not isinstance(
            valid_set, torch.utils.data.DataLoader
        ):
            valid_set = create_dataloader(valid_set, **valid_loader_kwargs)

        self._on_fit_start()

        for self.current_epoch in range(self.start_epoch, self.hparams["epochs"]):

            self._on_stage_start(Stage.TRAIN, self.current_epoch)
            self.modules.train()

            last_ckpt_time = time.time()

            t = self._get_iterable(
                train_set,
                desc=f"Epoch {self.current_epoch} train",
                initial=self.step,
                dynamic_ncols=True,
            )
            for batch in t:
                self.step += 1
                loss = self._fit_batch(batch)
                self.avg_train_loss = self._update_average(loss, self.avg_train_loss)
                self._update_progress(
                    t, self.step, stage="train", train_loss=self.avg_train_loss
                )

                if not (self.step % self.train_interval):
                    self._on_train_interval(self.current_epoch)

                if self.debug and self.step >= self.debug_batches:
                    break

                if (
                    self.checkpointer is not None
                    and self.ckpt_interval_minutes is not None
                    and time.time() - last_ckpt_time
                    >= self.ckpt_interval_minutes * 60.0
                ):
                    self._save_intra_epoch_ckpt()
                    last_ckpt_time = time.time()

            if not self.noprogressbar:
                t.close()

            # Run train "on_stage_end" on all processes
            self._on_stage_end(Stage.TRAIN, self.avg_train_loss, self.current_epoch)

            # Validation stage
            if valid_set is not None:
                self._on_stage_start(Stage.VALID, self.current_epoch)
                self.modules.eval()
                avg_valid_loss = 0.0
                valid_step = 0
                with torch.no_grad():
                    t = self._get_iterable(
                        valid_set,
                        desc=f"Epoch {self.current_epoch} valid",
                        dynamic_ncols=True,
                    )
                    for batch in t:
                        valid_step += 1
                        loss = self._evaluate_batch(batch, stage=Stage.VALID)
                        avg_valid_loss = self._update_average(loss, avg_valid_loss)
                        self._update_progress(
                            t, valid_step, stage="valid", valid_loss=avg_valid_loss
                        )

                    if not self.noprogressbar:
                        t.close()

                    # Only run validation "on_stage_end" on main process
                    self._on_stage_end(Stage.VALID, avg_valid_loss, self.current_epoch)
            self.step = 0
            self.avg_train_loss = 0.0
        self._on_fit_end()

    def _get_iterable(self, dataset: torch.utils.data.DataLoader, **kwargs):
        if not self.noprogressbar:
            from tqdm import tqdm

            t = tqdm(dataset, **kwargs)
            return t
        else:
            logger.warning(f"Running {self.__class__.__name__} without tqdm")
            return dataset

    def _update_progress(self, iterable: Iterable, step: int, stage: Stage, **kwargs):
        if self.noprogressbar:
            if not step % self.train_interval or (step >= len(iterable)) or (step == 1):
                self._update_progress_console(iterable, step, stage=stage, **kwargs)
        else:
            self._update_progress_tqdm(iterable, **kwargs)

    def _update_progress_tqdm(self, iterable: Iterable, **kwargs):
        if "metrics" in self.hparams:
            iterable.set_postfix(
                **kwargs,
                **self.hparams["metrics"].get_metrics(mode="mean"),
            )
        else:
            iterable.set_postfix(**kwargs)

    def _update_progress_console(
        self, iterable: Iterable, step: int, stage: Stage, **kwargs
    ):
        kwargs_string = ", ".join([f"{k}={v:.4f}" for k, v in kwargs.items()])
        if "metrics" in self.hparams:
            metrics = self.hparams["metrics"].get_metrics(mode="mean")
            metrics_string = ",".join([f"{k}={v:.4f}" for k, v in metrics.items()])
        else:
            metrics_string = ""

        print(
            f"Epoch {self.current_epoch} {stage}: step {step}/{len(iterable)}"
            f" -> {kwargs_string}, {metrics_string}"
        )
